/* $Id: subchannel.c 7619 2013-04-26 10:46:25Z chris $ */

#include <assert.h>
#include <string.h>

#include "subchannel.h"

extern const uint16_t crctbl[256];

void extract_subchannel(uint8_t *in, unsigned subchannel, uint8_t* out)
{
  unsigned bitno;
  uint8_t byte;
  int shift = 7 - subchannel;

  for (bitno = 0; bitno < 96; bitno++) {
    byte = (byte << 1) |
      (*(in++) >> shift) & 0x1;

    if (bitno % 8 == 7)
      *(out++) = byte;
  }
}

int parse_qsub(uint8_t *raw, struct qsub *qsub)
{
  uint16_t crc;
  int i;

  for (crc = 0, i = 0; i < 10; i++) {
    crc = crctbl[(crc >> 8) ^ raw[i]] ^ (crc << 8);
  }
  crc =~ crc;

  if (crc != (raw[10] << 8 | raw[11]))
    return E_INVALID_CRC;

  qsub->control = (raw[0] >> 4) & 0xf;
  qsub->mode = (raw[0] >> 0) & 0xf;

#define BCD_PAIR(x_) ((((x_) >> 4) & 0xf) * 10 + ((x_) & 0xf))
#define BCD_CHAR(x_) ((x_) + '0')
#define ALN_CHAR(x_) ((x_) < 10 ? (x_) + '0' : (x_) - 17 + 'A')

  /* TODO: Code elsewhere relies on the timing data we return being sane.  We therefore
     range check all the timing fields that we return.  However, we could do better.  We
     should range check the invidual BCD digits (as well as the final value), and the
     perform better checking of the textual data that we return (i.e. that it is actually
     alphanumeric). */
  switch (qsub->mode) {
  case QSUB_MODE_TIMING:
    qsub->u.timing.t.track = BCD_PAIR(raw[1]);
    qsub->u.timing.t.index = BCD_PAIR(raw[2]);
    qsub->u.timing.t.rel.m = BCD_PAIR(raw[3]);
    qsub->u.timing.t.rel.s = BCD_PAIR(raw[4]);
    qsub->u.timing.t.rel.f = BCD_PAIR(raw[5]);
    qsub->u.timing.t.abs.m = BCD_PAIR(raw[7]);
    qsub->u.timing.t.abs.s = BCD_PAIR(raw[8]);
    qsub->u.timing.t.abs.f = BCD_PAIR(raw[9]);

    if ((qsub->u.timing.t.track > 99 &&
         /* Quick hack to permit lead-out track */
         qsub->u.timing.t.track != BCD_PAIR(0xaa)) ||
        qsub->u.timing.t.index > 99 ||
        qsub->u.timing.t.rel.m > 99 ||
        qsub->u.timing.t.rel.s > 59 ||
        qsub->u.timing.t.rel.f > 74 ||
        qsub->u.timing.t.abs.m > 99 ||
        qsub->u.timing.t.abs.s > 59 ||
        qsub->u.timing.t.abs.f > 74)
      return E_INVALID_DATA;

    break;

  case QSUB_MODE_CATALOG:
    qsub->u.catalog.catalog[ 0] = BCD_CHAR((raw[1] & 0xf0) >> 4);
    qsub->u.catalog.catalog[ 1] = BCD_CHAR((raw[1] & 0x0f) >> 0);
    qsub->u.catalog.catalog[ 2] = BCD_CHAR((raw[2] & 0xf0) >> 4);
    qsub->u.catalog.catalog[ 3] = BCD_CHAR((raw[2] & 0x0f) >> 0);
    qsub->u.catalog.catalog[ 4] = BCD_CHAR((raw[3] & 0xf0) >> 4);
    qsub->u.catalog.catalog[ 5] = BCD_CHAR((raw[3] & 0x0f) >> 0);
    qsub->u.catalog.catalog[ 6] = BCD_CHAR((raw[4] & 0xf0) >> 4);
    qsub->u.catalog.catalog[ 7] = BCD_CHAR((raw[4] & 0x0f) >> 0);
    qsub->u.catalog.catalog[ 8] = BCD_CHAR((raw[5] & 0xf0) >> 4);
    qsub->u.catalog.catalog[ 9] = BCD_CHAR((raw[5] & 0x0f) >> 0);
    qsub->u.catalog.catalog[10] = BCD_CHAR((raw[6] & 0xf0) >> 4);
    qsub->u.catalog.catalog[11] = BCD_CHAR((raw[6] & 0x0f) >> 0);
    qsub->u.catalog.catalog[12] = BCD_CHAR((raw[7] & 0xf0) >> 4);
    qsub->u.catalog.frame       = BCD_PAIR(raw[9]);

    if (qsub->u.catalog.frame > 74)
      return E_INVALID_DATA;

    break;

  case QSUB_MODE_ISRC:
    qsub->u.isrc.isrc[ 0] = ALN_CHAR(((raw[1] & 0xfc) >> 2));
    qsub->u.isrc.isrc[ 1] = ALN_CHAR(((raw[1] & 0x03) << 4) | ((raw[2] & 0xf0) >> 4));
    qsub->u.isrc.isrc[ 2] = ALN_CHAR(((raw[2] & 0x0f) << 2) | ((raw[3] & 0xc0) >> 6));
    qsub->u.isrc.isrc[ 3] = ALN_CHAR(((raw[3] & 0x3f) << 0));
    qsub->u.isrc.isrc[ 4] = ALN_CHAR(((raw[4] & 0xfc) >> 2));
    qsub->u.isrc.isrc[ 5] = BCD_CHAR((raw[5] & 0xf0) >> 4);
    qsub->u.isrc.isrc[ 6] = BCD_CHAR((raw[5] & 0x0f) >> 0);
    qsub->u.isrc.isrc[ 7] = BCD_CHAR((raw[6] & 0xf0) >> 4);
    qsub->u.isrc.isrc[ 8] = BCD_CHAR((raw[6] & 0x0f) >> 0);
    qsub->u.isrc.isrc[ 9] = BCD_CHAR((raw[7] & 0xf0) >> 4);
    qsub->u.isrc.isrc[10] = BCD_CHAR((raw[7] & 0x0f) >> 0);
    qsub->u.isrc.isrc[11] = BCD_CHAR((raw[8] & 0xf0) >> 4);
    qsub->u.isrc.frame    = BCD_PAIR(raw[9]);

    if (qsub->u.isrc.frame > 74)
      return E_INVALID_DATA;

    break;

  default:
    /* NOTHING */
    break;
  }

#undef BCD_CHAR
#undef ALN_CHAR

  return 0;
}

static void add_to_msf(struct msf *msf, int frames)
{
  int m, s, f = frames;

  s = f / 75;
  f = f % 75;

  msf->f += f;
  if (msf->f >= 75) {
    msf->f -= 75;
    s++;
  }


  m = s / 60;
  s = s % 60;

  msf->s += s;
  if (msf->s >= 60) {
    msf->s -= 60;
    m++;
  }


  msf->m += m;

  assert(msf->f < 75);
  assert(msf->s < 60);
}

void add_to_timing(struct timing *timing, int frames)
{
  add_to_msf(&timing->rel, frames);
  add_to_msf(&timing->abs, frames);
}

void advance_timing(struct timing *timing, uint8_t abs_f)
{
  int delta = (int) abs_f - (int) timing->abs.f;
  if (delta < 0)
    delta += 75;

  add_to_timing(timing, delta);

  assert(timing->abs.f == abs_f);
}
